#version 440 core

//#define LOCAL_SIZE 32 //set on CPU side.

layout(local_size_x = LOCAL_SIZE) in;

// 16 KB at least supported by directx 11 GPUs (4096*sizeof(float) = 16 KB)
//#define SHARED_MEM_SIZE 4096  //set on CPU side.

shared float shared_mem[SHARED_MEM_SIZE];

layout(std430, binding = 0)
buffer Data {
    float data[]; // also SHARED_MEM_SIZE
};

float pow2(float x) { return x * x; }

void main()
{
    uint thread_id = gl_GlobalInvocationID.x;

    // Compute
    for(uint i = thread_id; i < SHARED_MEM_SIZE; i += LOCAL_SIZE)
    {
        float v = shared_mem[i];
        float diffSum = 0.0;

        
        for(uint j=0; j < SHARED_MEM_SIZE; j++){
            // version 1 broadcast
            diffSum += pow2(v- shared_mem[j]);
            // version 2 individual banks
            diffSum += pow2(v- shared_mem[(j + thread_id)% SHARED_MEM_SIZE]);
        }
        
        for(uint j=0; j< SHARED_MEM_SIZE; j+=2){
            // version 3 every second bank (2 elements at once)
            diffSum += pow2(v- shared_mem[(j+ thread_id*2)% SHARED_MEM_SIZE]);// uses only even banks
            diffSum += pow2(v- shared_mem[(j+ thread_id*2+1)% SHARED_MEM_SIZE]);// uses only uneven banks
        }

        for(uint j=0; j < SHARED_MEM_SIZE; j+=4){

            // version 4 every fourth bank (4 elements at once)
            diffSum += pow2(v- shared_mem[(j+ thread_id*4)% SHARED_MEM_SIZE]);// uses only banks 0, 4, 8, 12, ...
            diffSum += pow2(v- shared_mem[(j+ thread_id*4+1)% SHARED_MEM_SIZE]);// uses only banks 1, 5, 9, 13, ...
            diffSum += pow2(v- shared_mem[(j+ thread_id*4+2)% SHARED_MEM_SIZE]);// uses only banks 2, 6, 10, 14, ...
            diffSum += pow2(v- shared_mem[(j+ thread_id*4+3)% SHARED_MEM_SIZE]);// uses only banks 3, 7, 11, 15, ...
        }
        

        // generalized with stride
        //uint stride=8;
        for(uint j=0; j< SHARED_MEM_SIZE; j+= STRIDE){
            for(uint s=0; s< STRIDE; s++)
                diffSum += pow2(v- shared_mem[(j+ thread_id* STRIDE+ s)% SHARED_MEM_SIZE]);
        }
        
        // write result back to global memory
        data[i] = diffSum / (SHARED_MEM_SIZE - 1); // average squared difference
    }
}