#version 440 core

#define LOCAL_SIZE 64
#define STRIDE 2
#define BASE_INDEX 0

layout(local_size_x = LOCAL_SIZE) in;

// 16 KB at least supported by directx 11 GPUs (4096*sizeof(float) = 16 KB)
#define SHARED_MEM_SIZE 4096

shared float shared_mem[SHARED_MEM_SIZE];

layout(std430, binding = 0)
buffer Data {
    float data[]; // also SHARED_MEM_SIZE
};

float pow2(float x) { return x * x; }

void main()
{
    uint thread_id = gl_GlobalInvocationID.x;

    // Compute
    for(uint j = 0; j < 1; ++j)
    {
        for(uint i = thread_id; i < SHARED_MEM_SIZE; i += LOCAL_SIZE)
        {
            float v = shared_mem[i];
            float diffSum = 0.0;

            // generalized with stride
            //uint stride=8;
            for(uint j=BASE_INDEX; j< SHARED_MEM_SIZE + BASE_INDEX; j+= STRIDE){
                for(uint s=0; s< STRIDE; s++)
                    diffSum += pow2(v- shared_mem[(j+ thread_id* STRIDE+ s)% SHARED_MEM_SIZE]);
            }

            // write result back to global memory
            data[i] = diffSum / (SHARED_MEM_SIZE - 1); // average squared difference
        }
    }
}
